/*
 * netlist.c
 *
 * 
 * Author: 
 *  Richard Hult <rhult@hem.passagen.se>
 * 
 *  http://www.dtek.chalmers.se/~d4hult/oregano/ 
 * 
 * Copyright (C) 1999,2000  Richard Hult 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */
#include <config.h>
#include <gnome.h>
#include <stdio.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <locale.h>
#include "main.h"
#include "schematic.h"
#include "sim-settings.h"
#include "sheet-private.h"
#include "node-store.h"
#include "sheet-node.h"
#include "wire.h"
#include "part-private.h"
#include "part-property.h"
#include "netlist.h"

typedef struct {
	gint node_nr;
	GHashTable *pins;
	GHashTable *models;
	GSList *gnd_list;
	GSList *mark_list;
} NetlistData;

typedef struct {
	gint node_nr;
	gchar *name;
} Marker;

static void node_traverse (SheetNode *node, NetlistData *data);
static void wire_traverse (Wire *wire, NetlistData *data);

static void
node_foreach_reset (gpointer key, gpointer value, gpointer user_data)
{
	SheetNode *node = value;

	sheet_node_set_visited (node, FALSE);
}

static void
wire_traverse (Wire *wire, NetlistData *data)
{
	GSList *nodes;

	g_return_if_fail (wire != NULL);
	g_return_if_fail (IS_WIRE (wire));

	if (wire_is_visited (wire))
		return;

	wire_set_visited (wire, TRUE);

	for (nodes = wire_get_nodes (wire); nodes; nodes = nodes->next) {
		SheetNode *node = nodes->data;

		node_traverse (node, data);
	}
}

static void
node_traverse (SheetNode *node, NetlistData *data)
{
	GSList *wires, *pins;
	gchar *prop;

	g_return_if_fail (node != NULL);
	g_return_if_fail (IS_SHEET_NODE (node));

	if (sheet_node_is_visited (node))
		return;

	sheet_node_set_visited (node, TRUE);

	for (pins = node->pins; pins; pins = pins->next) {
		Pin *pin = pins->data;

		/*
		 * First see if the pin is an internal, special pin.
		 */
		prop = part_get_property (pin->part, "internal");
		if (prop) {
			if (!g_strcasecmp (prop, "marker")) {
				Marker *marker;
				gchar *name, *value;

				name = part_get_property (pin->part, "name");

				if (!name) {
					g_free (prop);
					continue;
				}
				
				value = part_property_expand_macros (pin->part, name);
				g_free (name);

				if (!value)
					continue;
				
				marker = g_new0 (Marker, 1);
				marker->node_nr = data->node_nr;
				marker->name = value;
				data->mark_list = g_slist_prepend (data->mark_list, marker);
			} else if (!g_strcasecmp (prop, "ground")) {
				data->gnd_list = g_slist_prepend (data->gnd_list, GINT_TO_POINTER (data->node_nr));
			}
			g_free (prop);
			continue;
		}

		/*
		 * Keep track of models to include. Need to be freed when the
		 * hash table is no longer needed.
		 */
		prop = part_get_property (pin->part, "model");
		if (prop) {
			if (!g_hash_table_lookup (data->models, prop))
				g_hash_table_insert (data->models, prop, NULL);
		}
		
		g_hash_table_insert (data->pins, pin, GINT_TO_POINTER (data->node_nr));
	}

	for (wires = node->wires; wires; wires = wires->next) {
		Wire *wire = wires->data;
		wire_traverse (wire, data);
	}
}

static void
node_foreach_traverse (gpointer key, gpointer value, NetlistData *data)
{
	SheetNode *node = value;

	/*
	 * Only visit nodes that are not already visited.
	 */
	if (!sheet_node_is_visited (node))
		node_traverse (node, data);
	else
		return;

	data->node_nr++;
}

static gint
compare_marker (gconstpointer a, gconstpointer b)
{
	const Marker *ma;
	gint node_nr;

	ma = a;
	node_nr = GPOINTER_TO_INT (b);

	if (ma->node_nr == node_nr)
		return 0;
	else
		return 1;
}

static void
foreach_model_write (char *model, gpointer user_data, FILE *f)
{
	char *str, *tmp1, *tmp2;

	tmp1 = g_concat_dir_and_file (OREGANO_LIBRARYDIR, model);
	tmp2 = g_strconcat (tmp1, ".mod", NULL);
	str = g_strdup_printf (".include %s\n", tmp2);
	fputs (str, f);

	g_free (tmp1);
	g_free (tmp2);
	g_free (str);
}

static void
foreach_model_free (gpointer key, char *model, gpointer user_data)
{
	g_free (key);
}

static char *
linebreak (char *str)
{
	char **split, *tmp;
	GString *out;
	int i;

	split = g_strsplit (str, "\\", 0);	

	out = g_string_new ("");

	i = 0;
	while (split[i] != NULL) {
		if (split[i][0] == 'n') {
			if (strlen (split[i]) > 1) {
				out = g_string_append_c (out, '\n');
				out = g_string_append (out, split[i] + 1);
			}
		} else {
			out = g_string_append (out, split[i]);
		}
		
		i++;
	}

	g_strfreev (split);
	tmp = out->str;
	g_string_free (out, FALSE); /* Don't free the string data. */
	return tmp;
}

gchar *
nl_generate (Schematic *sm, gchar *name)
{
	NetlistData *data;
	GList *parts, *wires;
	Part *part;
	gint pin_nr, num_pins, num_nodes, num_gnd_nodes, i, j;
	Pin *pins;
	gchar *template, **template_split;
	NodeStore *store;
	gchar **node2real;
	FILE *f;
	char *tmp_filename = NULL;
	gchar *filename;
	gint fd;
	gchar *old_locale;
	SimSettings *sim_settings;

	schematic_log_clear (sm); 

	if (name == NULL) {
		do {
			if (tmp_filename != NULL)
				free (tmp_filename);
			
			tmp_filename = tempnam (NULL, "oreg");
			fd = open (tmp_filename, O_CREAT | O_EXCL | O_TRUNC
				   | O_RDWR, 0600);
			
		} while (fd == -1);
		
		filename = g_strdup (tmp_filename);
		free (tmp_filename);
		
		f = fdopen (fd, "w");
	} else {
		filename = g_strdup (name);
		f = fopen (name, "w");
	}

	if (f == NULL) {
		g_free (filename);
		return NULL;
	}

	old_locale = g_strdup (setlocale (LC_NUMERIC, NULL));
	setlocale (LC_NUMERIC, "C");

	{
		char *title;
		title = schematic_get_filename (sm);
		fputs (title ? title : "Title: <unset>", f);
		fputs ("\n", f);
		fputs (".OPTIONS ITL1=500 ITL4=100 GMIN=1e-20\n", f); /* Needed for the 555 model. */
	}

	store = schematic_get_store (sm);

	node_store_node_foreach (store, (GHFunc *) node_foreach_reset, NULL);
	for (wires = store->wires; wires; wires = wires->next) {
		Wire *wire = wires->data;
		wire_set_visited (wire, FALSE);
	}

	data = g_new0 (NetlistData, 1);
	data->pins = g_hash_table_new (g_direct_hash, g_direct_equal);
	data->models = g_hash_table_new (g_str_hash, g_str_equal);
	data->node_nr = 1;
	data->gnd_list = NULL;
	data->mark_list = NULL;

	node_store_node_foreach (store, (GHFunc *) node_foreach_traverse, data);

	num_gnd_nodes = g_slist_length (data->gnd_list);

	if (num_gnd_nodes == 0) {
		schematic_log_append (sm, _("No ground node. Aborting.\n"));
		schematic_log_show (sm); 

		fclose (f);
		unlink (filename);
		g_free (filename);
		filename = NULL;
		goto bail_out;
	}		

	num_nodes = data->node_nr - 1;

	/*
	 * Build an array for going from node nr to "real node nr",
	 * where gnd nodes are nr 0 and the rest of the nodes are
	 * 1,2, 3, ...
	 */
	node2real = g_new0 (gchar*, num_nodes + 1);

	for (i = 1, j = 1; i <= num_nodes; i++) {
		GSList *mlist;

		if (g_slist_find (data->gnd_list, GINT_TO_POINTER (i))) {
			node2real[i] = g_strdup ("0");
		} else if ((mlist = g_slist_find_custom (data->mark_list, GINT_TO_POINTER (i), compare_marker))) {
			Marker *marker = mlist->data;
			node2real[i] = g_strdup(marker->name);
		} else {
			node2real[i] = g_strdup_printf ("%d", j++);
		}
	}
	
	for (parts = store->parts; parts; parts = parts->next) {
		gchar *tmp, *internal;

		part = parts->data;

		internal = part_get_property (part, "internal");
		if (internal) {
			g_free (internal);
			continue;
		}

		tmp = part_get_property (part, "template");
		if (!tmp)
			continue;
		template = part_property_expand_macros (part, tmp);
		g_free (tmp);

		tmp = linebreak (template);
		g_free (template);
		template = tmp;

		num_pins = part_get_num_pins (part);
		pins = part_get_pins (part);

		template_split = g_strsplit (template, " ", 0);
		g_free (template);
		template = "";

		i = 0;
		while (template_split[i] != NULL && template_split[i][0] != '%') {
			template = g_strconcat (template, template_split[i++], " ", NULL);
		}

		for (pin_nr = 0; pin_nr < num_pins; pin_nr++) {
			gint node_nr;

			node_nr = GPOINTER_TO_INT (g_hash_table_lookup (data->pins, &pins[pin_nr]));
			if (!node_nr) {
				g_warning ("\nCouln't find part, pin_nr %d!\n", pin_nr);
			} else {
				gchar *tmp;
				tmp = node2real[node_nr];
				template = g_strconcat (template, tmp, " ", NULL);
				i++;
			}

			while (template_split[i] != NULL) {
				if (template_split[i][0] == '%')
					break;
				template = g_strconcat (template, template_split[i++], " ", NULL);
			}
		}

		while (template_split[i] != NULL) {
			template = g_strconcat (template, template_split[i++], " ", NULL);
		}

		g_strfreev (template_split);
		fputs (template, f);
		fputs ("\n", f);
	}

	for (i = 0; i < num_nodes + 1; i++)
		g_free (node2real[i]);

	g_free (node2real);

	fputs ("\n", f);

	g_hash_table_foreach (data->models, (GHFunc)foreach_model_write, f);

	fputs ("\n", f);

	/*
	 * Make sure we get ascii output.
	 * FIXME: Use binary to speed things up and simplify parsing?
	 */
	fputs (".control\n", f);
	fputs ("set filetype=ascii\n", f);
	fputs (".endc\n", f);

	/* FIXME: move this to simulate.c */
	sim_settings = schematic_get_sim_settings (sm);
	if (sim_settings_get_trans (sim_settings)) {
		gchar *tmp;
		gdouble start, stop, step;
		gboolean step_enable;

	 	start = sim_settings_get_trans_start (sim_settings);
 		stop = sim_settings_get_trans_stop (sim_settings);
 		step = sim_settings_get_trans_step (sim_settings);
		step_enable = sim_settings_get_trans_step_enable (sim_settings);

		if (step_enable)
			tmp = g_strdup_printf (".tran 1e-3 %g %g %g\n", stop, start, step);
		else
			tmp = g_strdup_printf (".tran 1e-3 %g %g\n", stop, start);

		fputs (tmp, f);
		g_free (tmp);
	}

 	fputs ("\n.END\n", f);

 	fclose (f);

 bail_out:
	g_hash_table_foreach (data->models, (GHFunc)foreach_model_free, NULL);
	g_hash_table_destroy (data->models);

	g_hash_table_destroy (data->pins);
	g_free (data);

	setlocale (LC_NUMERIC, old_locale);
	g_free (old_locale);

	return filename;
}

