/*
 * plot.c
 *
 * Author: 
 *  Richard Hult <rhult@hem.passagen.se>
 * 
 *  http://www.dtek.chalmers.se/~d4hult/oregano/ 
 * 
 * Copyright (C) 1999,2000  Richard Hult 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <math.h>
#include <gnome.h>
#include <glade/glade.h>
#include "main.h"
#include "schematic.h"
#include "sim-engine.h"
#include "dialogs.h"
#include "plot.h"
#include "smallicon.h"
#include "pixmaps/mini_icon_plot.xpm"

#define PLOT_PADDING_X 40
#define PLOT_PADDING_Y 40
#define PLOT_WIDTH 400
#define PLOT_HEIGHT 300
#define DEBUG_PLOT 0

#define PLOT_AXIS_COLOR "black"
#define PLOT_CURVE_COLOR "medium sea green"

static gdouble x_min, x_max;

typedef struct {
	GtkWidget *window;
	GtkWidget *canvas;

	GnomeCanvasGroup *plot_group;   /* holds the actual curve */
	GnomeCanvasGroup *pad_group;    /* contains axes, plot title etc */ 
	GnomeCanvasGroup *axis_group;   /* the x/y-axes */

	SimEngine *sim;

	gchar *title;
	gchar *xtitle;
	gchar *ytitle;

	gint width;           /* width of the plot, excluding */
	gint height;          /* axes, titles and padding etc */

	gdouble x_factor;
	gdouble y_factor;

	gdouble plot_min, plot_max;

	gdouble padding_x;     /* padding around the plot. Note that */
	gdouble padding_y;     /* titles, axes etc live here */

	gint selected;         /* the currently selected plot in the clist */
	gint prev_selected;
} Plot;

static GtkWidget *plot_window_create (Plot *plot);
static void make_plot (Plot *plot, gint plot_number);
static void destroy_window (GtkWidget *widget, Plot *plot);
static void destroy_plot (Plot *plot);
static gint delete_event_cb (GtkWidget *widget, GdkEvent *event, Plot *plot);
static void getbins(gdouble, gdouble, gint,  gdouble *, gdouble *, gint *, gdouble *);

/* Define menus and toolbar */
static GnomeUIInfo plot_file_menu [] = {
	
	{ GNOME_APP_UI_ITEM, N_("_Close"), NULL, destroy_window, NULL, NULL,
	  GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_CLOSE, 'w', GDK_CONTROL_MASK 
	},
	GNOMEUIINFO_END
};

static GnomeUIInfo plot_plot_menu [] = {
	{ GNOME_APP_UI_ITEM, N_("_Preferences..."), NULL, NULL, NULL, NULL,
	  GNOME_APP_PIXMAP_STOCK, GNOME_STOCK_MENU_PREF, 0, 0 },
	GNOMEUIINFO_END
};

static GnomeUIInfo plot_help_menu[] =
{
/*	GNOMEUIINFO_HELP ("Schematic Plot"),*/
	GNOMEUIINFO_MENU_ABOUT_ITEM(NULL, NULL),
	GNOMEUIINFO_END
};

GnomeUIInfo plot_main_menu[] = {
	GNOMEUIINFO_SUBTREE(N_("_File"), plot_file_menu),
	GNOMEUIINFO_SUBTREE(N_("_Plot"), plot_plot_menu),
	GNOMEUIINFO_MENU_HELP_TREE (plot_help_menu),
	GNOMEUIINFO_END
};

static gint
delete_event_cb (GtkWidget *widget, GdkEvent *event, Plot *plot)
{
	plot->window = NULL;
	return FALSE; /* yes, please destroy me */
}

/* Call this to close the plot window */
static void
destroy_window (GtkWidget *widget, Plot *plot)
{
	gtk_widget_destroy (plot->window);
	plot->window = NULL;
}

static void 
show_plot (Plot *plot)
{
	if (plot)
		destroy_plot (plot);

	make_plot (plot, plot->selected);
}

static void
plot_selected (GtkCList *clist, gint row, gint col, GdkEvent *event, Plot *plot)
{
        plot->prev_selected = plot->selected;
        plot->selected = row + 1;
	show_plot (plot);
}

void destroy_plot (Plot *plot)
{
	if (plot->axis_group) gtk_object_destroy (GTK_OBJECT (plot->axis_group));
	if (plot->plot_group) gtk_object_destroy (GTK_OBJECT (plot->plot_group));

	plot->axis_group = NULL;
	plot->plot_group = NULL;
}

void make_plot (Plot *plot, gint plot_number)
{
	GnomeCanvas *canvas;
	GnomeCanvasGroup *plot_group, *axis_group;
	GnomeCanvasItem *plot_title, *dmy;
	GnomeCanvasPoints *points;
	gchar *xtitle, *ytitle, *title;
	gchar *text;

	double plot_title_width, lbl;
	double y_factor, y, y1;
	double x_factor, x, x1;
	double y_min, y_max;
	gint i;	
	GtkArg arg;

	gint Ndiv[2] = { 10, 5 }, nBins[2];
	gdouble Wmn[2], Wmx[2], Bw[2];
	const static double eps = 0.00001;
	static char line[64];
	char *ysc, *xsc;
	gdouble sx,sy;
	
	y_min = plot->sim->transient.min_data[plot_number];
	y_max = plot->sim->transient.max_data[plot_number];

	/* FIXME! hack... make sure this works */
	if (y_min > 0) {
		plot->plot_min = 0;
	} else {
		plot->plot_min = y_min;
	}
	
	if (y_max < 0) {
		plot->plot_max = 0;
	} else {
		plot->plot_max = y_max;
	}

//---
	sprintf (line, "%.2g", plot->plot_max);
	ysc = strchr (line, 'e');
	if (ysc) {
		ysc = g_strdup (ysc + 1);
		sscanf (ysc, "%lf", &sy);
		sy = pow (10, sy);
	}
//--- 

	x_min = 0;
	x_max = plot->sim->transient.sim_length;

//---
	sprintf (line, "%.2g", x_max);
	xsc = strchr (line, 'e');
	if (xsc) {
		xsc = g_strdup (xsc + 1);
		sscanf (xsc, "%lf", &sx);
		sx = pow (10, sx);
	}
//---

	/* used to go from actual data values to canvas coordinates */
	y_factor = (PLOT_HEIGHT) / (plot->plot_max - plot->plot_min);
	x_factor = (PLOT_WIDTH) / (x_max - x_min) ;

	plot->x_factor = x_factor;
	plot->y_factor = y_factor;
		
#if DEBUG_PLOT	
	g_print ("\n");
	g_print ("x_min %.2g, x_max %.2g\n", x_min, x_max);
	g_print ("y_min %.2g, y_max %.2g\n", y_min, y_max);

/*
	g_print ("Scale factor y: %g\n", y_factor);
	g_print ("Scale factor x: %g\n", x_factor);
*/
#endif

	canvas = GNOME_CANVAS (plot->canvas);
	canvas->close_enough = 5;

	xtitle = plot->xtitle;
	ytitle = plot->ytitle;
	title = plot->title;

	/*
	 * 1. Make a group to hold the actual plot. Lets us keep track of the plot to
	 *    destroy it later. This lets us use relative coordinates too, which makes
	 *    the padding code transparent and simple.
	 */
	plot_group = GNOME_CANVAS_GROUP (gnome_canvas_item_new (gnome_canvas_root (canvas),
								gnome_canvas_group_get_type (),
								"x", plot->padding_x,
								"y", plot->padding_y,
								NULL));
	plot->plot_group = plot_group;

#if DEBUG_PLOT
	g_print ("plot height: %d, plot_min: %g,  y_factor: %g\n", PLOT_HEIGHT, plot->plot_min, y_factor);
#endif
	
	/*
	 * 2. Plot the curve, using the calculated scale factors. 
	 */
	points = gnome_canvas_points_new (plot->sim->transient.num_data-1);
	for (i = 0; i < plot->sim->transient.num_data-1; i++){
		double data;

		data = g_array_index (plot->sim->transient.data[plot_number], double, i);
		y = PLOT_HEIGHT - ((data - plot->plot_min) * y_factor);
		points->coords[2 * i + 1] = y;

		data = g_array_index (plot->sim->transient.data[0], double, i);
		points->coords[2 * i] = data * x_factor;
	}
	
	gnome_canvas_item_new (plot_group,
			       gnome_canvas_line_get_type (),
			       "points", points,
			       "fill_color", PLOT_CURVE_COLOR,
			       NULL);
	gnome_canvas_points_free (points);

#if 0
	gtk_signal_connect (GTK_OBJECT (plot_group), 
			    "event",
			    GTK_SIGNAL_FUNC (plot_event),
			    plot);
#endif
		
	/*
	 * 3. Make a group to hold the the x- and y-axis. 
	 */
	axis_group = GNOME_CANVAS_GROUP (gnome_canvas_item_new (gnome_canvas_root (canvas),
								gnome_canvas_group_get_type (),
								"x", plot->padding_x,
								"y", plot->padding_y,
								NULL));
	plot->axis_group = axis_group;
	
	/*
	 * 4.1 Start with the the x-axis. Put it at y = 0.
	 */
	y = PLOT_HEIGHT - ((0.0  - plot->plot_min) * y_factor);

#if 0
	g_print ("plot_min=%g\n", plot_min);
	g_print ("plot_max=%g\n", plot_max);
#endif	

	points = gnome_canvas_points_new (2);
	points->coords[0] = 0.0;
	points->coords[1] = y;
	points->coords[2] = PLOT_WIDTH + 10.0;    /* extra 10.0 !!! */
	points->coords[3] = y;

	gnome_canvas_item_new (axis_group,
			       gnome_canvas_line_get_type (),
			       "points", points,
			       "fill_color", PLOT_AXIS_COLOR,
			       "last_arrowhead", TRUE,
			       "arrow_shape_a", 6.0,
			       "arrow_shape_b", 8.0,
			       "arrow_shape_c", 5.0,
			       NULL);
	
	gnome_canvas_points_free (points);

//---	
	/* x-axis title */
	if ( xsc ) 
		sprintf(line,"%s (x%.0g)",xtitle,sx);
	else 
		sprintf(line,"%s",xtitle);
//---

	y = PLOT_HEIGHT - ((0.0  - plot->plot_min) * y_factor);
	plot_title = gnome_canvas_item_new (axis_group,
					    gnome_canvas_text_get_type (),
					    "text", xtitle,
					    "x", (double)PLOT_WIDTH,
					    "y", y + 5.0,
					    "fill_color", PLOT_AXIS_COLOR,
					    "anchor", GTK_ANCHOR_NE,
					    "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
					    NULL);
	
	/* Right justify the x-axis title */
	/*	arg.type = GTK_TYPE_DOUBLE;
		arg.name = "text_width";
		gtk_object_getv(GTK_OBJECT(plot_title), 1, &arg);
		plot_title_width = GTK_VALUE_DOUBLE(arg);
		gtk_object_set (GTK_OBJECT(plot_title), "x", PLOT_WIDTH - plot_title_width, NULL);	
	*/

//---
	arg.type = GTK_TYPE_DOUBLE;
	arg.name = "text_height";
	gtk_object_getv(GTK_OBJECT(plot_title), 1, &arg);
	lbl = GTK_VALUE_DOUBLE(arg);
	gtk_object_set (GTK_OBJECT(plot_title), "y",PLOT_HEIGHT + 1.25*lbl, NULL);	
//---	

	/* 4.2 y-axis, at x = 0.0.
	 *
	 */
	points = gnome_canvas_points_new (2);
	points->coords[0] = 0.0;
	points->coords[1] = PLOT_HEIGHT;
	points->coords[2] = 0.0;
	points->coords[3] = 0.0 - 10.0;   /* extra 10.0 !! */

	gnome_canvas_item_new (axis_group,
			       gnome_canvas_line_get_type (),
			       "points", points,
			       "fill_color", PLOT_AXIS_COLOR,
			       "last_arrowhead", TRUE,
			       "arrow_shape_a", 6.0, /* vert inner */
			       "arrow_shape_b", 8.0, /* vert */
			       "arrow_shape_c", 5.0,  /* horiz */
			       NULL);
	
	/* y-axis title */
	plot_title = gnome_canvas_item_new (axis_group,
					    gnome_canvas_text_get_type (),
					    "text", ytitle,
					    "x", 0.0 - 5.0,
					    "y", 0.0,
					    "fill_color", PLOT_AXIS_COLOR,
					    "anchor", GTK_ANCHOR_NE,
					    "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
					    NULL);
	
	text = g_strdup_printf ("%3.3g", plot->plot_max);
	gnome_canvas_item_new (axis_group,
			       gnome_canvas_text_get_type (),
			       "text", text,
			       "x", 5.0,
			       "y", -15.0,
			       "fill_color", "red",
			       "anchor", GTK_ANCHOR_NW,
			       "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
			       NULL);
	g_free (text);

	text = g_strdup_printf ("%3.3g", plot->plot_min);
	gnome_canvas_item_new (axis_group,
			       gnome_canvas_text_get_type (),
			       "text", text,
			       "x", 5.0,
			       "y", (double)PLOT_HEIGHT + 15.0,
			       "fill_color", "red",
			       "anchor", GTK_ANCHOR_NW,
			       "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
			       NULL);
	g_free (text);
	
	/*
	 * 4.3 put tics on the x- and y-axis.
	 */
	getbins (x_min, x_max, Ndiv[0], Wmn, Wmx, nBins, Bw);
	if ((x_min - Wmn[0]) > eps) {
		Wmn[0] += Bw[0]; 
		nBins[0]--;
	}
	if ((Wmx[1] - x_max) > eps) {
		Wmx[0] -= Bw[0];
		nBins[0]--;
	}
	getbins (0, Bw[0], Ndiv[1], Wmn + 1,Wmx + 1, nBins + 1, Bw + 1);

	for (x = Wmn[0]; x <= Wmx[0]; x += Bw[0]) {
		points->coords[0] = x * x_factor;
		points->coords[1] = y;
		points->coords[2] = points->coords[0];
		points->coords[3] = y - 5.0;
		sprintf(line,"%.3g", x / (xsc ? sx : 1));
		gnome_canvas_item_new (axis_group,
				       gnome_canvas_line_get_type (),
				       "points", points,
				       "fill_color", PLOT_AXIS_COLOR,
				       NULL);
		dmy = gnome_canvas_item_new (axis_group,
					     gnome_canvas_text_get_type (),
					     "text", line,
					     "x",points->coords[0],
					     "y", y,
					     "fill_color", "red",
					     "anchor", GTK_ANCHOR_NW,
					     "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
					     NULL);
		arg.type = GTK_TYPE_DOUBLE;
		arg.name = "text_width";
		gtk_object_getv(GTK_OBJECT(dmy), 1, &arg);
		lbl = GTK_VALUE_DOUBLE(arg);
		gtk_object_set (GTK_OBJECT(dmy), "x",points->coords[0]-lbl/2., NULL);	
		arg.type = GTK_TYPE_DOUBLE;
		arg.name = "text_height";
		gtk_object_getv(GTK_OBJECT(dmy), 1, &arg);
		lbl = GTK_VALUE_DOUBLE(arg);
		gtk_object_set (GTK_OBJECT(dmy), "y",y+0.5*lbl, NULL);  
		
		
		for (x1=Wmn[1]+Bw[1];x1<Wmx[1]+eps && x+x1<x_max;x1+=Bw[1]) {
			points->coords[0] = (x+x1)*x_factor;
			points->coords[1] = y;
			points->coords[2] = points->coords[0];
			points->coords[3] = y-2.0;
			gnome_canvas_item_new (axis_group,
					       gnome_canvas_line_get_type (),
					       "points", points,
					       "fill_color", PLOT_AXIS_COLOR,
					       NULL); 	 
		}
	}
	
	
	x = 0.0;
	y = 0.0;
	getbins(plot->plot_min,plot->plot_max,Ndiv[0],Wmn,Wmx,nBins,Bw);
	if ((plot->plot_min-Wmn[0]) > eps) { Wmn[0] += Bw[0]; nBins[0]--; }
	if ((Wmx[1]-plot->plot_max) > eps) { Wmx[0] -= Bw[0]; nBins[0]--; }
	getbins(0.,Bw[0],Ndiv[1],Wmn+1,Wmx+1,nBins+1,Bw+1);
	
	for ( y=Wmn[0];y<=Wmx[0];y+=Bw[0]) {
		points->coords[0] = x;
		points->coords[1] = PLOT_HEIGHT - (y-plot->plot_min)*y_factor;
		points->coords[2] = x+5.;
		points->coords[3] = points->coords[1];
		sprintf(line,"%.3g",y/(ysc ? sy : 1.));
		gnome_canvas_item_new (axis_group,
				       gnome_canvas_line_get_type (),
				       "points", points,
				       "fill_color", PLOT_AXIS_COLOR,
				       NULL);
		dmy = gnome_canvas_item_new (axis_group,
					     gnome_canvas_text_get_type (),
					     "text", line,
					     "x",points->coords[0],
					     "y",points->coords[1],
					     "fill_color", "red",
					     "anchor", GTK_ANCHOR_NW,
					     "font", "-adobe-helvetica-medium-r-normal-*-12-*-*-*-*-*-*-*",
					     NULL);
		
		arg.type = GTK_TYPE_DOUBLE;
		arg.name = "text_width";
		gtk_object_getv(GTK_OBJECT(dmy), 1, &arg);
		lbl = GTK_VALUE_DOUBLE(arg);
		gtk_object_set (GTK_OBJECT(dmy), "x",points->coords[0]-lbl-4, NULL);	
		arg.type = GTK_TYPE_DOUBLE;
		arg.name = "text_height";
		gtk_object_getv(GTK_OBJECT(dmy), 1, &arg);
		lbl = GTK_VALUE_DOUBLE(arg);
		gtk_object_set (GTK_OBJECT(dmy), "y",points->coords[1]-lbl/2., NULL);  
		
		
		for (y1=Wmn[1]+Bw[1];y1<Wmx[1]+eps && y+y1<plot->plot_max;y1+=Bw[1]) {
			points->coords[0] = x;
			points->coords[1] = PLOT_HEIGHT - ((y+y1)-plot->plot_min) * y_factor;
			points->coords[2] = x+2.;
			points->coords[3] = points->coords[1];
			gnome_canvas_item_new (axis_group,
					       gnome_canvas_line_get_type (),
					       "points", points,
					       "fill_color", PLOT_AXIS_COLOR,
					       NULL); 	 
		}
	}
	//---
						   
						   
#if 0
	x = 0.0;
	y = PLOT_HEIGHT - ((0.0  - plot->plot_min) * y_factor);
	while (x < PLOT_WIDTH){
		points->coords[0] = x;
		points->coords[1] = y + 3.0;
		points->coords[2] = x;
		points->coords[3] = y + 0.0;
		
		gnome_canvas_item_new (axis_group,
				       gnome_canvas_line_get_type (),
				       "points", points,
				       "fill_color", PLOT_AXIS_COLOR,
				       NULL);
		x += 15.0;
	}

	x = 0.0;
	y = 0.0;
	while (y < PLOT_HEIGHT){
		points->coords[0] = x;
		points->coords[1] = PLOT_HEIGHT - y;
		points->coords[2] = x - 3.0;
		points->coords[3] = PLOT_HEIGHT - y;
		
		gnome_canvas_item_new (axis_group,
				       gnome_canvas_line_get_type (),
				       "points", points,
				       "fill_color", PLOT_AXIS_COLOR,
				       NULL);
		y += 15.0;
	}
#endif
	
	gnome_canvas_points_free (points);
	
	plot_title = gnome_canvas_item_new (axis_group,
					    gnome_canvas_text_get_type (),
					    "text", title,
					    "x", 0.0,
					    "y", 0.0 - plot->padding_y / 2,
					    "fill_color", PLOT_AXIS_COLOR,
					    "anchor", GTK_ANCHOR_W,
					    "font", "-adobe-helvetica-bold-r-normal-*-12-*-*-*-*-*-*-*",
					    NULL);
	
	/* Now find out the width of the plot title and make it centered */
	arg.type = GTK_TYPE_DOUBLE;
	arg.name = "text_width";
	gtk_object_getv(GTK_OBJECT(plot_title), 1, &arg);
	plot_title_width = GTK_VALUE_DOUBLE(arg);
	gtk_object_set (GTK_OBJECT(plot_title), "x", PLOT_WIDTH / 2 - plot_title_width / 2, NULL);	

//---
	if ( ysc ) g_free(ysc);
	if ( xsc ) g_free(xsc);
//---
}

static GtkWidget *
plot_window_create (Plot *plot)
{  
	GtkWidget *clist;
	GtkWidget *canvas, *outer_table, *window, *button;
	GtkStyle *style;
	GdkColormap *colormap;
	GdkPixmap *icon;
	GdkBitmap *icon_mask;
	GladeXML *gui;
	gchar *msg;
	
	window = gnome_app_new ("plot", "Oregano - Plot window");
/*	gnome_app_create_menus_with_data (GNOME_APP (window), plot_main_menu, plot);
 */	
	gtk_signal_connect (GTK_OBJECT (window), "delete_event",
			    GTK_SIGNAL_FUNC (delete_event_cb), plot);
	
	if (!g_file_exists (OREGANO_GLADEDIR "/plot-window.glade")) {
		msg = g_strdup_printf (_("Could not find the required file:\n%s"), 
				       OREGANO_GLADEDIR "/plot-window.glade");
		oregano_error (msg);
		g_free (msg);
		return NULL;
	}

	gui = glade_xml_new (OREGANO_GLADEDIR "/plot-window.glade", "plot-table");
	if (!gui) {
		oregano_error (_("Could not create plot window."));
		return NULL;
	}

	outer_table = glade_xml_get_widget (gui, "plot-table");
	canvas = plot->canvas = glade_xml_get_widget (gui, "plot-canvas");

	gnome_app_set_contents (GNOME_APP (window), outer_table);

	button = glade_xml_get_widget (gui, "close-button");
	gtk_signal_connect (GTK_OBJECT (button), "clicked",
			    GTK_SIGNAL_FUNC (destroy_window), plot);

	gtk_widget_set_usize (canvas, PLOT_WIDTH + PLOT_PADDING_X*2, PLOT_HEIGHT + PLOT_PADDING_Y*2);
	gnome_canvas_set_scroll_region (GNOME_CANVAS (canvas), 0, 0, 
					PLOT_WIDTH + PLOT_PADDING_X*2, 
					PLOT_HEIGHT + PLOT_PADDING_Y*2);


	clist = glade_xml_get_widget (gui, "variable-list");
	gtk_clist_set_column_min_width (GTK_CLIST (clist), 0, 90);
	gtk_object_set_data (GTK_OBJECT (window), "clist", clist);

   	style =  gtk_style_new ();
	colormap = gtk_widget_get_colormap (GTK_WIDGET (canvas));
	gdk_color_white (colormap, &style->bg[GTK_STATE_NORMAL]);
	gtk_widget_set_style (GTK_WIDGET (canvas), style);

	gtk_widget_realize (GTK_WIDGET (window));
	icon = gdk_pixmap_create_from_xpm_d (window->window, &icon_mask, NULL, mini_icon_plot_xpm);
	set_small_icon (window->window, icon, icon_mask);
	
	gtk_widget_show_all (window);  
	return window;
}

int
plot_show (SimEngine *engine)
{
	GtkWidget *clist;
	Plot *plot;
	int i;

	g_return_val_if_fail (engine != NULL, FALSE);
	g_return_val_if_fail (IS_SIM_ENGINE (engine), FALSE);

	plot = g_new0 (Plot, 1);

	plot->sim = engine;
	plot->window = plot_window_create (plot);

	plot->title = g_strdup ("Plot - Transient Analysis");
	plot->xtitle = g_strdup ("time");
	plot->ytitle = g_strdup ("volts");

	plot->padding_x = 40;
	plot->padding_y = 40;

	plot->axis_group = NULL;
	plot->plot_group = NULL;

       	clist = gtk_object_get_data (GTK_OBJECT (plot->window), "clist");

	gtk_clist_freeze (GTK_CLIST (clist));
	for (i = 1; i < engine->transient.num_var; i++) {
		if (strchr (engine->transient.var_names[i], '#') == NULL)
			gtk_clist_append (GTK_CLIST (clist), &engine->transient.var_names[i]);
	}
	gtk_clist_thaw (GTK_CLIST (clist));

	gtk_signal_connect (GTK_OBJECT (clist), "select_row",
			    GTK_SIGNAL_FUNC (plot_selected), plot);

	gtk_clist_select_row (GTK_CLIST (clist), 0, 0);
	
	return TRUE;
}


/*
 * GetBins
 *
 *   Author: Carlos
 *   Purpose: Finds out the binning for the scale of the control
 *            Wmin,Wmax       : range of the scale
 *            Nold            : wanted number of bins
 *            BinLow, BinHigh : Optimized low and hig bins for which
 *                              a label is to be drawn
 *            Nbins           : number of bins
 *            Bwidth          : bin width between ticks
 *
 */
void getbins (gdouble Wmin, gdouble Wmax, gint Nold,
	      gdouble *BinLow, gdouble *BinHigh, gint *Nbins, gdouble *Bwidth)
{
	gint lwid, kwid;
	gint ntemp = 0;
	gint jlog  = 0;
	gdouble siground = 0;
	gdouble alb, awidth, sigfig,atest;
	
	gdouble AL = MIN (Wmin, Wmax);
	gdouble AH = MAX (Wmin, Wmax);
	if (AL == AH) 
		AH = AL + 1;
	
/*
 * If Nold  ==  -1, program uses binwidth input from calling routine.
 */
	if (Nold == -1 && (*Bwidth) > 0) 
		goto L90;
	
	ntemp = MAX (Nold, 2);
	if (ntemp < 1) 
		ntemp = 1;
	
/*
 * Get nominal bin width in exponential form.
 */
	
 L20:
	awidth = (AH - AL) / ((gdouble)ntemp);
	if (awidth > 1.e30) 
		goto LOK;
	
	jlog   = (gint)(log10 (awidth));
	if (awidth <= 1) 
		jlog--;
	
	sigfig = awidth * ((gdouble)pow (10, -jlog));
	
/*
 * Round mantissa up to 1, 2, 2.5, 5, or 10. 
 */
	
	if (sigfig <= 1)
		siground = 1;
	else if (sigfig <= 2)
		siground = 2;
//   else if (sigfig <= 2.5)  siground = 2.5;
	else if (sigfig <= 5)
		siground = 5;
	else {
		siground = 1;   
		jlog++; 
	}
	
	(*Bwidth) = siground * pow (10, jlog);
	
/*
 * Get new bounds from new width Bwidth.
 */
	
 L90:
	alb = AL / (*Bwidth);
	lwid = (gint) (alb);
	if (alb < 0)
		lwid--;
	(*BinLow) = siground*((gdouble)lwid)*pow(10,jlog);
	alb = AH / (*Bwidth) + 1.00001f;
	kwid = (gint) alb;
	if (alb < 0) 
		kwid--;
	(*BinHigh) = siground * ((gdouble) kwid) * pow (10, jlog);
	(*Nbins) = kwid - lwid;
	if (Nold == -1)
		goto LOK;
	if (Nold <= 5) {          /* Request for one bin is difficult case. */
		if (Nold > 1 || (*Nbins) == 1)
			goto LOK;
		
		(*Bwidth) = (*Bwidth) * 2;
		(*Nbins) = 1;
		goto LOK;
	}
	if (2 * (*Nbins) == Nold) {
		ntemp++;
		goto L20;
	}
	
 LOK:
	atest = (*Bwidth) * 0.0001f;
	if (fabs ((*BinLow) - Wmin)  >= atest) {
		(*BinLow) += (*Bwidth);
		(*Nbins)--;
	}
	
	if (fabs ((*BinHigh) - Wmax) >= atest) { 
		(*BinHigh) -= (*Bwidth);
		(*Nbins)--;
	}
}

