/*
 * sim-engine.c
 *
 * Simulate engine class, interface to SPICE.
 * 
 * Author: 
 *  Richard Hult <rhult@hem.passagen.se>
 * 
 *  http://www.dtek.chalmers.se/~d4hult/oregano/ 
 * 
 * Copyright (C) 1999,2000  Richard Hult 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <gnome.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <sys/types.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <sys/stat.h>

#include "main.h" /* get rid of this... */
#include "sim-settings.h"
#include "sim-engine.h"
#include "schematic.h"

#define READ 0
#define WRITE 1

static void sim_engine_class_init (SimEngineClass *klass);
static void sim_engine_init (SimEngine *engine);

static void data_input_cb (SimEngine *engine, gint source, GdkInputCondition condition);
static void error_input_cb (SimEngine *engine, gint source, GdkInputCondition condition);
static void transient_input_cb (SimEngine *engine, gint source, GdkInputCondition condition);
static void op_input_cb (SimEngine *engine, gint source, GdkInputCondition condition);

enum {
	DONE,
	ABORTED,
	CANCELLED,
	LAST_SIGNAL
};

static guint sim_engine_signals [LAST_SIGNAL] = { 0 };
static GtkObject *parent_class = NULL;

guint
sim_engine_get_type (void)
{
	static guint sim_engine_type = 0;

	if (!sim_engine_type) {
		static const GtkTypeInfo sim_engine_info = {
			"SimEngine",
			sizeof (SimEngine),
			sizeof (SimEngineClass),
			(GtkClassInitFunc) sim_engine_class_init,
			(GtkObjectInitFunc) sim_engine_init,
			/* reserved_1 */ NULL,
			/* reserved_2 */ NULL,
			(GtkClassInitFunc) NULL,
		};
		
		sim_engine_type = gtk_type_unique (GTK_TYPE_OBJECT, &sim_engine_info);
	}

	return sim_engine_type;
}

static void
sim_engine_destroy (GtkObject *object)
{
	SimEngine *engine;
	int i, num;

	engine = SIM_ENGINE (object);

	num = engine->transient.num_var;

	g_free (engine->transient.var_names);
	g_free (engine->transient.var_units);

	for (i = 0; i < num; i++) {
		if (engine->transient.data[i])
			g_array_free (engine->transient.data[i], FALSE);
	}	
	g_free (engine->transient.data);

	g_free (engine->transient.min_data);
	g_free (engine->transient.max_data);

	/*
	 * Chain up. 
	 */
	GTK_OBJECT_CLASS (parent_class)->destroy (object);
}

static void
sim_engine_class_init (SimEngineClass *klass)
{
	GtkObjectClass *object_class;

	object_class = (GtkObjectClass *)klass;
	parent_class = gtk_type_class (GTK_TYPE_OBJECT);

	object_class->destroy = sim_engine_destroy;

	sim_engine_signals [DONE] = 
		gtk_signal_new ("done",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (SimEngineClass, done),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);

	sim_engine_signals [ABORTED] = 
		gtk_signal_new ("aborted",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (SimEngineClass, aborted),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);

	sim_engine_signals [CANCELLED] = 
		gtk_signal_new ("cancelled",
				GTK_RUN_FIRST,
				object_class->type,
				GTK_SIGNAL_OFFSET (SimEngineClass, cancelled),
				gtk_marshal_NONE__NONE,
				GTK_TYPE_NONE, 0);

	gtk_object_class_add_signals (object_class, sim_engine_signals, LAST_SIGNAL);
}

static void
sim_engine_init (SimEngine *engine)
{
	engine->running = FALSE;
	engine->progress = 0.0;

	engine->child_pid = 0;

	engine->data_input_id = 0;
	engine->error_input_id = 0;

	engine->inputfp = NULL;
	engine->outputfp = NULL;
	engine->errorfp = NULL;
	
	engine->has_warnings = FALSE;
	engine->has_errors = FALSE;
}

SimEngine *
sim_engine_new (Schematic *sm)
{
	SimEngine *engine;

	engine = gtk_type_new (sim_engine_get_type ());

	engine->sm = sm;
	engine->sim_settings = schematic_get_sim_settings (sm);

	return engine;
}

void
sim_engine_start_with_file (SimEngine *engine, const gchar *netlist)
{
	int netlist_fd;

	if (pipe (engine->to_child) < 0)
		g_error ("Error creating stdin pipe");
	if (pipe (engine->to_parent) < 0)
		g_error ("Error creating stdout pipe");
	if (pipe (engine->to_parent_error) < 0)
		g_error ("Error creating stderr pipe");
	
	engine->child_pid = fork ();
	if (engine->child_pid == 0) {
		gchar *spice = oregano.simexec;
		gchar *args[4] = { oregano.simexec, "-s", "-n", NULL };

		/* The child executes here. */
		netlist_fd = open (netlist, O_RDONLY);
		if (netlist_fd == -1)
			g_error ("Error opening netlist.");
		
		close (engine->to_child[WRITE]);
		close (engine->to_parent[READ]);
		close (engine->to_parent_error[READ]);

		dup2 (netlist_fd, STDIN_FILENO);
		close (netlist_fd);
		close (engine->to_child[READ]);

		/* Map stderr and stdout to their pipes. */
		dup2 (engine->to_parent[WRITE], STDOUT_FILENO);
		dup2 (engine->to_parent_error[WRITE], STDERR_FILENO);
		close (engine->to_parent[WRITE]);
		close (engine->to_parent_error[WRITE]);
		
		/* Execute spice. */
		execvp (spice, args);

		/* We should never get here. */
		g_warning ("Error executing spice.");
		_exit (1);
	}
	
	if (engine->child_pid == -1) {
		g_warning ("Could not fork child process.");
		return;
	}

	/* The parent executes here. */

	schematic_log_clear (engine->sm);

	/* Set up communication. */
	close (engine->to_child[READ]);
	close (engine->to_parent[WRITE]);
	close (engine->to_parent_error[WRITE]);
	
	engine->errorfp = fdopen (engine->to_parent_error[READ], "r");
	engine->inputfp = fdopen (engine->to_parent[READ], "r");
	engine->outputfp = fdopen (engine->to_child[WRITE], "w");
	if (engine->inputfp == NULL || engine->outputfp == NULL) {
		g_warning ("Error executing child process");
		return;
	}

	engine->running = TRUE;
	engine->data_input_id = gdk_input_add (
		engine->to_parent[READ],
		GDK_INPUT_READ,
		(GdkInputFunction) data_input_cb,
		(gpointer) engine);

	engine->error_input_id = gdk_input_add (
		engine->to_parent_error[READ],
		GDK_INPUT_READ,
		(GdkInputFunction) error_input_cb,
		(gpointer) engine);	
}

void
sim_engine_stop (SimEngine *engine)
{
	int status;

	g_return_if_fail (engine != NULL);
	g_return_if_fail (IS_SIM_ENGINE (engine));

	if (!engine->running)
		return;

	gdk_input_remove (engine->data_input_id);
	gdk_input_remove (engine->error_input_id);
	engine->data_input_id = 0;
	engine->error_input_id = 0;
	
	kill (engine->child_pid, SIGKILL);
	waitpid (engine->child_pid, &status, WUNTRACED);
	engine->child_pid = 0;
	
	fclose (engine->errorfp);
	fclose (engine->inputfp);
	fclose (engine->outputfp);
	
	engine->running = FALSE;
}

static void
data_input_cb (SimEngine *engine, gint source, GdkInputCondition condition)
{
	int status;
	static gchar buf[256];

	if (!(condition & GDK_INPUT_READ))
		return;

	if (fgets (buf, 255, engine->inputfp) == NULL){
		engine->running = FALSE;

		gdk_input_remove (engine->data_input_id);
		gdk_input_remove (engine->error_input_id);

		kill (engine->child_pid, SIGKILL);
		waitpid (engine->child_pid, &status, WUNTRACED);
		
		fclose (engine->errorfp);
		fclose (engine->inputfp);
		fclose (engine->outputfp);

		gtk_signal_emit_by_name (GTK_OBJECT (engine), "aborted");

		return;
	}
	
	if (!strncmp (buf, "Plotname: ", strlen ("Plotname: "))){
		int i = strlen ("Plotname: ");
		
		if (!strncmp (buf + i, "Transient Analysis", strlen ("Transient Analysis"))){
			double start, stop;

			gdk_input_remove (engine->data_input_id);

			start = sim_settings_get_trans_start (engine->sim_settings);
			stop = sim_settings_get_trans_stop (engine->sim_settings);

			engine->transient.sim_length = stop - start;
			/* FIXME: This is not yet used. */
			engine->transient.step_size = 5e-3;

			engine->transient.state = 0;
			engine->data_input_id = gdk_input_add (
				engine->to_parent[READ],
				GDK_INPUT_READ,
				(GdkInputFunction) transient_input_cb,
				(gpointer) engine);

			return;
		} else if (!strncmp (buf + i, "Operating Point", strlen ("Operating Point"))){

			gdk_input_remove (engine->data_input_id);

			engine->op.state = 0;
			engine->data_input_id = gdk_input_add (
				engine->to_parent[READ],
				GDK_INPUT_READ,
				(GdkInputFunction) op_input_cb,
				(gpointer) engine);

			return;
		}
	}
}

static void
error_input_cb (SimEngine *engine, gint source, GdkInputCondition condition)
{
	static gchar buf[256]; 

	if (!(condition & GDK_INPUT_READ))
		return;

	/*
	 * Store the error messages.
	 */
	if (fgets (buf, 255, engine->errorfp) != NULL)
		if (strncmp (buf, "@@@", 3) != 0) {
			schematic_log_append (engine->sm, buf); 
			engine->has_warnings = TRUE;
		}
}

static void
transient_input_cb (SimEngine *engine, gint source, GdkInputCondition condition)
{
	static gchar buf[256];
	gchar **tmp = NULL;
	double value;
	int i;

	if (!(condition & GDK_INPUT_READ))
		return;

	if (fgets (buf, 255, engine->inputfp) == NULL) {
		int status;

		engine->running = FALSE;

		gdk_input_remove (engine->data_input_id);
		gdk_input_remove (engine->error_input_id);

		kill (engine->child_pid, SIGKILL);
		waitpid (engine->child_pid, &status, WUNTRACED);
		
		fclose (engine->errorfp);
		fclose (engine->inputfp);
		fclose (engine->outputfp);

		if (engine->transient.state != 4) {
			engine->has_errors = TRUE;
			gtk_signal_emit_by_name (GTK_OBJECT (engine), "aborted");
		} else
			gtk_signal_emit_by_name (GTK_OBJECT (engine), "done");

		return;
	}	

	if (!strncmp (buf, "Title", strlen ("Title"))) {
		gdk_input_remove (engine->data_input_id);
		engine->data_input_id = gdk_input_add (
			engine->to_parent[READ],
			GDK_INPUT_READ,
			(GdkInputFunction) data_input_cb,
			(gpointer) engine);
		return;
	}
	
	switch (engine->transient.state){
	case 0:
		/* Skip until we get the number of variables. */
		if (!strncmp (buf, "No. Variables: ", strlen ("No. Variables: "))) {
			int n =	atoi (buf + strlen ("No. Variables: "));
			engine->transient.num_var = n;
			engine->transient.got_var = 0;
			engine->transient.var_names = (char**) g_new0 (gpointer, n);
			engine->transient.var_units = (char**) g_new0 (gpointer, n);

			engine->transient.data = (GArray **) g_new0 (gpointer, n);
			for (i = 0; i < n; i++) {
				engine->transient.data[i] = g_array_new (TRUE, TRUE, sizeof (double));
			}

			engine->transient.min_data = g_new (double, n);
			engine->transient.max_data = g_new (double, n);
			for (i = 0; i < n; i++) {
				engine->transient.min_data[i] = G_MAXDOUBLE;
				engine->transient.max_data[i] = -G_MAXDOUBLE;
			}

			/* Ready for the next step in the parsing pipeline. */
			engine->transient.state++;
		}
		return;

	case 1:
		/* Skip until we find the variable declarations. */
		if (!strncmp (buf, "Variables:", strlen ("Variables:"))) {
			engine->transient.state++;
		}
		return;

	case 2:
		/* Get the variable names. */
		tmp = g_strsplit (buf, "\t", 0);
		engine->transient.var_names[engine->transient.got_var] = g_strdup (tmp[2]);
		engine->transient.var_units[engine->transient.got_var] = g_strdup (tmp[3]);
		g_strfreev (tmp);

		if ((engine->transient.got_var + 1) == engine->transient.num_var)
			engine->transient.state++;
		else
			engine->transient.got_var++;
		return;

	case 3:
		/* Skip until we find simulation data. */
		if (!strncmp (buf, "Values:", strlen ("Values:"))) {
			engine->transient.state++;
		}

		engine->transient.got_var = 0;
		engine->transient.num_data = 0;
		return;

	case 4:
		/* Parse data for the first variable. */
		tmp = g_strsplit (buf, "\t", 0);
		/* The time seems to be delimited with two tabs... */
		value = g_strtod (tmp[2], NULL);
		engine->progress = value / engine->transient.sim_length;
		if (engine->progress > 1.0)
			engine->progress = 1.0;
		engine->transient.data[0] = g_array_append_val (engine->transient.data[0], value);
		g_strfreev (tmp);

		engine->transient.state++;
		engine->transient.got_var++;
		engine->transient.num_data++;
		break;

	case 5:
		/* Parse data for variables other than the first. */
		value = g_strtod (buf, NULL);
		engine->transient.data[engine->transient.got_var] = 
			g_array_append_val (engine->transient.data[engine->transient.got_var], value);

		/* Update the minimal and maximal values so far. */
		if (value < engine->transient.min_data[engine->transient.got_var])
			engine->transient.min_data[engine->transient.got_var] = value;
		if (value > engine->transient.max_data[engine->transient.got_var])
			engine->transient.max_data[engine->transient.got_var] = value;

		if ((engine->transient.got_var + 1) == engine->transient.num_var) {
			engine->transient.got_var = 0;
			engine->transient.state--;
		} else
			engine->transient.got_var++;
			
		break;
	}
}

static void
op_input_cb (SimEngine *engine, gint source, GdkInputCondition condition)
{
	static gchar buf[256];
	gchar **tmp = NULL;
	double value;
	int i;

	g_print ("op\n");

	if (!(condition & GDK_INPUT_READ))
		return;

	if (fgets (buf, 255, engine->inputfp) == NULL) {
		int status;

		engine->running = FALSE;

		gdk_input_remove (engine->data_input_id);
		gdk_input_remove (engine->error_input_id);

		kill (engine->child_pid, SIGKILL);
		waitpid (engine->child_pid, &status, WUNTRACED);
		
		fclose (engine->errorfp);
		fclose (engine->inputfp);
		fclose (engine->outputfp);

		if (engine->op.state != 4) {
			gtk_signal_emit_by_name (GTK_OBJECT (engine), "aborted");
		} else
			gtk_signal_emit_by_name (GTK_OBJECT (engine), "done");

		return;
	}	

	if (!strncmp (buf, "Title", strlen ("Title"))) {
		gdk_input_remove (engine->data_input_id);
		engine->data_input_id = gdk_input_add (
			engine->to_parent[READ],
			GDK_INPUT_READ,
			(GdkInputFunction) data_input_cb,
			(gpointer) engine);
		return;
	}
		
	switch (engine->op.state){
	case 0:
		/* Skip until we get the number of variables. */
		if (!strncmp (buf, "No. Variables: ", strlen ("No. Variables: "))) {
			int n =	atoi (buf + strlen ("No. Variables: "));
			engine->op.num_var = n;
			engine->op.got_var = 0;
			engine->op.var_names = g_new0 (gpointer, n);
			engine->op.var_units = g_new0 (gpointer, n);

			engine->op.data = g_new0 (char, n);
			for (i = 0; i < n; i++) {
				engine->op.data[i] = g_array_new (TRUE, TRUE, sizeof (double));
			}

			/* Ready for the next step in the parsing pipeline. */
			engine->op.state++;
		}
		return;

	case 1:
		/* Skip until we find the variable declarations. */
		if (!strncmp (buf, "Variables:", strlen ("Variables:"))) {
			engine->op.state++;
		}
		return;

	case 2:
		/* Get the variable names. */
		tmp = g_strsplit (buf, "\t", 0);
		engine->op.var_names[engine->op.got_var] = g_strdup (tmp[2]);
		engine->op.var_units[engine->op.got_var] = g_strdup (tmp[3]);
		g_strfreev (tmp);

		if ((engine->op.got_var + 1) == engine->op.num_var)
			engine->op.state++;
		else
			engine->op.got_var++;
		return;

	case 3:
		/* Skip until we find simulation data. */
		if (!strncmp (buf, "Values:", strlen ("Values:"))) {
			engine->op.state++;
		}

		engine->op.got_var = 0;
		engine->op.num_data = 0;
		return;

	case 4:

		/* Parse data for the first variable. */
		tmp = g_strsplit (buf, "\t", 0);
		/* This seems to be delimited with two tabs... */
		value = g_strtod (tmp[2], NULL);
		engine->op.data[0] = g_array_append_val (engine->op.data[0], value);
		g_strfreev (tmp);

		engine->op.state++;
		engine->op.got_var++;
		engine->op.num_data++;
		break;

	case 5:

		/* Parse data for variables other than the first. */
		value = g_strtod (buf, NULL);
		engine->op.data[engine->op.got_var] = 
			g_array_append_val (engine->op.data[engine->op.got_var], value);

		if ((engine->op.got_var + 1) == engine->op.num_var) {
			engine->op.got_var = 0;
			engine->op.state--;
		} else
			engine->op.got_var++;
			
		break;
	}
}
