/*
 * sim-settings.c
 *
 * 
 * Author: 
 *  Richard Hult <rhult@hem.passagen.se>
 * 
 *  http://www.dtek.chalmers.se/~d4hult/oregano/ 
 * 
 * Copyright (C) 1999,2000  Richard Hult 
 * 
 * This program is free software; you can redistribute it and/or 
 * modify it under the terms of the GNU General Public License as 
 * published by the Free Software Foundation; either version 2 of the 
 * License, or (at your option) any later version. 
 * 
 * This program is distributed in the hope that it will be useful, 
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public
 * License along with this program; if not, write to the
 * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
 * Boston, MA 02111-1307, USA.
 */

#include <config.h>
#include <unistd.h>
#include <gnome.h>
#include <glade/glade.h>
#include "main.h"
#include "sim-settings.h"
#include "schematic.h"
#include "schematic-view.h"
#include "dialogs.h"
#include "oregano-utils.h"

struct _SimSettingsPriv {
	/* Transient analysis. */
	GtkWidget *w_trans_enable, *w_trans_start, *w_trans_stop, *w_trans_step, *w_trans_step_enable;
	gboolean trans_enable;
	gchar *trans_start;
	gchar *trans_stop;
	gchar *trans_step;
	gboolean trans_step_enable;

	/* Fourier analysis. Replace with something sane later. */
	GtkWidget *w_four_start, *w_four_stop, *w_four_enable;
	gchar *four_start;
	gchar *four_stop;
	gboolean four_enable;
};

static void 
trans_enable_cb (GtkWidget *widget, SimSettings *s)
{
	gboolean enable, step_enable;

	enable = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
	step_enable = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (s->priv->w_trans_step_enable));

	gtk_widget_set_sensitive (s->priv->w_trans_start, enable);
	gtk_widget_set_sensitive (s->priv->w_trans_stop, enable);
	gtk_widget_set_sensitive (s->priv->w_trans_step, enable & step_enable);
	gtk_widget_set_sensitive (s->priv->w_trans_step_enable, enable);
	
	gnome_property_box_changed (GNOME_PROPERTY_BOX (s->pbox));
}

static void 
trans_step_enable_cb (GtkWidget *widget, SimSettings *s)
{
	gboolean enable, step_enable;

	step_enable = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (widget));
	enable = gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (s->priv->w_trans_enable));

	gtk_widget_set_sensitive (s->priv->w_trans_step, step_enable & enable);
	
	gnome_property_box_changed (GNOME_PROPERTY_BOX (s->pbox));
}

static void 
entry_changed_cb (GtkWidget *widget, SimSettings *s)
{
	gnome_property_box_changed (GNOME_PROPERTY_BOX (s->pbox));
}

static void
apply_callback (GtkWidget *w, int page_nr, SimSettings *s)
{
	if (page_nr == -1 ) {
		if (s->priv->trans_start)
			g_free (s->priv->trans_start);
		s->priv->trans_start = gtk_editable_get_chars (GTK_EDITABLE (s->priv->w_trans_start), 0, -1);
		if (s->priv->trans_stop)
			g_free (s->priv->trans_stop);
		s->priv->trans_stop = gtk_editable_get_chars (GTK_EDITABLE (s->priv->w_trans_stop), 0, -1);
		if (s->priv->trans_step)
			g_free (s->priv->trans_step);
		s->priv->trans_step = gtk_editable_get_chars (GTK_EDITABLE (s->priv->w_trans_step), 0, -1);

		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (s->priv->w_trans_enable)))
			s->priv->trans_enable = TRUE;
		else
			s->priv->trans_enable = FALSE;

		if (gtk_toggle_button_get_active (GTK_TOGGLE_BUTTON (s->priv->w_trans_step_enable)))
			s->priv->trans_step_enable = TRUE;
		else
			s->priv->trans_step_enable = FALSE;
	}
}

static int 
close_callback (GtkWidget *widget, SimSettings *s)
{
	if (s->priv->trans_enable) {
		if (sim_settings_get_trans_start (s) >= sim_settings_get_trans_stop (s)) {
			oregano_error (_("Transient analysis:\n\n"
					 "Start time must be less than stop time."));
			return TRUE;
		}
	}

	s->pbox = NULL;
	return FALSE;
}

static int 
delete_event_cb (GtkWidget *widget, GdkEvent *event, gpointer data)
{
	return FALSE;
}

SimSettings *
sim_settings_new (Schematic *sm)
{
	SimSettings *s;

	s = g_new0 (SimSettings, 1);
	s->sm = sm;

	s->priv = g_new0 (SimSettingsPriv, 1);

	/*
	 * Set some default settings.
	 */
	s->priv->trans_enable = TRUE;
	s->priv->trans_start = g_strdup ("0 s");
	s->priv->trans_stop = g_strdup ("5 ms");
	s->priv->trans_step = g_strdup ("0.1 ms");
	s->priv->trans_step_enable = FALSE;

	s->priv->four_enable = FALSE;

	return s;
}

gboolean
sim_settings_get_trans (SimSettings *sim_settings)
{
	return sim_settings->priv->trans_enable;
}

gdouble
sim_settings_get_trans_start (SimSettings *sim_settings)
{
	gchar *text = sim_settings->priv->trans_start;
	return oregano_strtod (text, 's');
}

gdouble
sim_settings_get_trans_stop (SimSettings *sim_settings)
{
	gchar *text = sim_settings->priv->trans_stop;
	return oregano_strtod (text, 's');
}

gdouble
sim_settings_get_trans_step (SimSettings *sim_settings)
{
	gchar *text = sim_settings->priv->trans_step;
	return oregano_strtod (text, 's');
}

gdouble
sim_settings_get_trans_step_enable (SimSettings *sim_settings)
{
	return sim_settings->priv->trans_step_enable;
}

void
sim_settings_set_trans (SimSettings *sim_settings, gboolean enable)
{
	sim_settings->priv->trans_enable = enable;
}

void
sim_settings_set_trans_start (SimSettings *sim_settings, gchar *str)
{
	if (sim_settings->priv->trans_start)
		g_strdup (sim_settings->priv->trans_start);
	sim_settings->priv->trans_start = g_strdup (str);
}

void
sim_settings_set_trans_stop (SimSettings *sim_settings, gchar *str)
{
	if (sim_settings->priv->trans_stop)
		g_strdup (sim_settings->priv->trans_stop);
	sim_settings->priv->trans_stop = g_strdup (str);
}

void
sim_settings_set_trans_step (SimSettings *sim_settings, gchar *str)
{
	if (sim_settings->priv->trans_step)
		g_strdup (sim_settings->priv->trans_step);
	sim_settings->priv->trans_step = g_strdup (str);
}

void
sim_settings_set_trans_step_enable (SimSettings *sim_settings, gboolean enable)
{
	sim_settings->priv->trans_step_enable = enable;
}

void
sim_settings_show (GtkWidget *widget, SchematicView *sv)
{
	GtkWidget *w, *pbox;
	GladeXML *gui;
	SimSettings *s;
	Schematic *sm;

	g_return_if_fail (sv != NULL);

	sm = schematic_view_get_schematic (sv);
	s = schematic_get_sim_settings (sm);

	/* Only allow one instance of the property box per schematic. */
	if (s->pbox){
		gdk_window_raise (GTK_WIDGET (s->pbox)->window);
 		return;
	}		

	if (!g_file_exists (OREGANO_GLADEDIR "/sim-settings.glade")) {
		oregano_error (_("Could not create simulation settings dialog."));
		return;
	}

	gui = glade_xml_new (OREGANO_GLADEDIR "/sim-settings.glade", "toplevel");
	if (!gui) {
		oregano_error (_("Could not create simulation settings dialog."));
		return;
	}

	w = glade_xml_get_widget (gui, "toplevel");
	if (!w) {
		oregano_error (_("Could not create simulation settings dialog."));
		return;
	}

	pbox = w;
	s->pbox = GNOME_PROPERTY_BOX (pbox);

	gtk_signal_connect (GTK_OBJECT (pbox), "delete_event",
			    GTK_SIGNAL_FUNC (delete_event_cb), s);

	/* Setup callbacks. */
	gtk_signal_connect (GTK_OBJECT (pbox), "apply",
			    GTK_SIGNAL_FUNC (apply_callback), s);
	gtk_signal_connect (GTK_OBJECT (pbox), "close",
			    GTK_SIGNAL_FUNC (close_callback), s);

	w = glade_xml_get_widget (gui, "trans_start");
	if (s->priv->trans_start)
		gtk_entry_set_text (GTK_ENTRY (w), s->priv->trans_start);
	s->priv->w_trans_start = w;
	gtk_signal_connect (GTK_OBJECT (w), "changed",
			    GTK_SIGNAL_FUNC (entry_changed_cb), s);

	w = glade_xml_get_widget (gui, "trans_stop");
	if (s->priv->trans_stop)
		gtk_entry_set_text (GTK_ENTRY (w), s->priv->trans_stop);
	s->priv->w_trans_stop = w;
	gtk_signal_connect (GTK_OBJECT (w), "changed",
			    GTK_SIGNAL_FUNC (entry_changed_cb), s);

	w = glade_xml_get_widget (gui, "trans_step");
	if (s->priv->trans_step)
		gtk_entry_set_text (GTK_ENTRY (w), s->priv->trans_step);
	s->priv->w_trans_step = w;
	gtk_signal_connect (GTK_OBJECT (w), "changed",
			    GTK_SIGNAL_FUNC (entry_changed_cb), s);

	w = glade_xml_get_widget (gui, "trans_enable");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), s->priv->trans_enable);
	s->priv->w_trans_enable = w;

	w = glade_xml_get_widget (gui, "trans_step_enable");
	gtk_toggle_button_set_active (GTK_TOGGLE_BUTTON (w), s->priv->trans_step_enable);
	s->priv->w_trans_step_enable = w;

	gtk_signal_connect (GTK_OBJECT (s->priv->w_trans_enable), "clicked",
			    GTK_SIGNAL_FUNC (trans_enable_cb), s);

	gtk_signal_connect (GTK_OBJECT (s->priv->w_trans_step_enable), "clicked",
			    GTK_SIGNAL_FUNC (trans_step_enable_cb), s);


	trans_enable_cb (s->priv->w_trans_enable, s);
	trans_step_enable_cb (s->priv->w_trans_step_enable, s);
	
	return;
}


